from FreeBodyEngine.graphics.fbusl.ast_nodes import *
from FreeBodyEngine.graphics.fbusl.semantic import Function
from FreeBodyEngine.utils import abstractmethod
from typing import Literal


class Injector:
    """
    Injects into an FBUSL AST.
    """
    def __init__(self):
        self.tree: Tree = Tree()

    def init(self, shader_type: Literal['vert', 'frag'], file_path: str):
        self.shader_type = shader_type 
        self.file_path = file_path

    def get_builtins(self) -> dict[str, list]:
        return {}
        
    def pre_lexer_inject(self, source: str) -> str:
        "Modifies the raw shader source code text."
        return source

    def _pre_generation_inject(self, tree: Tree):
        self.tree = tree
        return self.pre_generation_inject()

    def pre_generation_inject(self) -> str:
        "Modifies the AST right before code generation."
        return self.tree

    def inject(self):
        return self.tree
    
    def find_main_function(self) -> FuncDecl | None:
        def recursive_search(node):
            if isinstance(node, FuncDecl) and node.name.name == "main":
                return node

            for value in vars(node).values():
                if isinstance(value, Node):
                    result = recursive_search(value)
                    if result:
                        return result
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, Node):
                            result = recursive_search(item)
                            if result:
                                return result
            return None

        for root in self.tree.children:
            result = recursive_search(root)
            if result:
                return result

        return None

    def replace_node(self, target: Node, replacement: Node):
        def recursive_replace(current: Node):
            for name, value in vars(current).items():
                if isinstance(value, Node):
                    if value is target:
                        setattr(current, name, replacement)
                    else:
                        recursive_replace(value)

                elif isinstance(value, list):
                    new_list = []
                    for item in value:
                        if isinstance(item, Node):
                            if item is target:
                                new_list.append(replacement)
                            else:
                                recursive_replace(item)
                                new_list.append(item)
                        else:
                            new_list.append(item)
                    setattr(current, name, new_list)

        for root in self.tree.children:
            recursive_replace(root)

    def find_parent(self, target: Node) -> Node | None:
        def recursive_search(current: Node):
            for name, value in vars(current).items():
                if isinstance(value, Node):
                    if value is target:
                        return current
                    result = recursive_search(value)
                    if result:
                        return result
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, Node):
                            if item is target:
                                return current
                            result = recursive_search(item)
                            if result:
                                return result
            return None

        for root in self.tree.children:
            result = recursive_search(root)
            if result:
                return result

        return None
    
    def find_nodes(self, attr_name, attr_val) -> list[Node]:
        nodes = []

        def recursive_search(node):
            if hasattr(node, attr_name) and getattr(node, attr_name) == attr_val:
                nodes.append(node)

            for name, value in vars(node).items():
                if isinstance(value, Node):
                    recursive_search(value)
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, Node):
                            recursive_search(item)



        for node in self.tree.children:
            recursive_search(node)

        return nodes



"""
The FreeBody Universal Shader Language is a Transpiled Shader Language.
It includes features present in many shader languages so it can
be easily compiled into them.
"""

import sys


class Position:
    def __init__(self, line: int = 0, file: str = None):
        self.line = line
        self.file = file

    def __repr__(self):
        return f"{self.line}"

def fbusl_error(msg, position: Position=Position()):
    f = position.file
    if position.file == None:
        f = "Unkown File"
    print(f'\033[91mFBUSL ERROR: {msg} in file "{f}", line {position.line}.\033[0m')
    sys.exit()


from FreeBodyEngine.graphics.fbusl import parser
from FreeBodyEngine.graphics.fbusl import ast_nodes
from FreeBodyEngine.graphics.fbusl import injector
from FreeBodyEngine.graphics.fbusl import semantic
from FreeBodyEngine.graphics.fbusl import generator
from FreeBodyEngine.graphics.fbusl.compiler import compile



__all__ = ["fbusl_error", "parser", "ast_nodes", "semantic", "generator", "injector", "compile"]


from typing import List
import re
from FreeBodyEngine.graphics.fbusl import fbusl_error, Position
from typing import List, Optional
from enum import Enum, auto
from FreeBodyEngine.graphics.fbusl.node import *
import sys


class Token:
    def __init__(self, kind: str, value, pos: Position = Position()):
        self.kind = kind
        self.value = value
        self.pos = pos

    def __repr__(self):
        return f"Token({self.kind}, {repr(self.value)})"


class TokenType(Enum):
    ARROW = auto()
    DECORATOR = auto()
    QUALIFIER = auto()
    KEYWORD = auto()
    IDENT = auto()
    FLOAT = auto()
    INT = auto()
    BOOL = auto()
    SYMBOL = auto()
    OPERATOR = auto()
    WHITESPACE = auto()
    NEWLINE = auto()
    COMMENT = auto()
    INDENT = auto()
    DEDENT = auto()


TOKEN_TYPES = [
    (TokenType.NEWLINE, r"\n"),  # newlines first
    (TokenType.WHITESPACE, r"[ \t]+"),  # only spaces and tabs
    (TokenType.ARROW, r"->"),
    (TokenType.DECORATOR, r"@(?:uniform|input|output|define)"),
    (TokenType.QUALIFIER, r"(low|med|flat|high)"),
    (TokenType.KEYWORD, r"\b(def|return|struct|if|buffer|elif|else|not|and|or)\b"),
    (TokenType.BOOL, r"(True|False)"),
    (TokenType.IDENT, r"[a-zA-Z_][a-zA-Z0-9_]*"),
    (TokenType.FLOAT, r"\d(?:_?\d)*\.\d(?:_?\d)*"),
    (TokenType.INT, r"\d(?:_?\d)*"),
    (TokenType.SYMBOL, r"[{}()[\]:,\.]"),
    (TokenType.OPERATOR, r"(\+=|-=|\*=|/=|=|\+|-|\*|/|==|!=|<=|>=|<|>)"),
    (TokenType.COMMENT, r"#.*"),
]

TOKEN_REGEX = [(kind, re.compile(pattern)) for kind, pattern in TOKEN_TYPES]


class Lexer:
    def __init__(self, code: str, filename: str = None):
        self.code = code
        self.filename = filename
        self.pos = 0
        self.line = 1
        self.length = len(code)
        self.indents = [0]  # stack of indentation levels
        self._pending_tokens: List[Token] = []

    def next_token(self) -> Optional[Token]:
        if self._pending_tokens:
            return self._pending_tokens.pop(0)

        if self.pos >= self.length:
            if len(self.indents) > 1:
                self.indents.pop()
                return Token(TokenType.DEDENT, "", Position(self.line, self.filename))
            return None

        if self.code[self.pos] == "\n":
            self.pos += 1
            self.line += 1
            start_pos = self.pos
            while self.pos < self.length and self.code[self.pos] in " \t":
                self.pos += 1
            indent = self.pos - start_pos
            last_indent = self.indents[-1]

            if indent > last_indent:
                self.indents.append(indent)

                self._pending_tokens.append(
                    Token(TokenType.INDENT, "", Position(self.line, self.filename))
                )
            elif indent < last_indent:
                while len(self.indents) > 1 and indent < self.indents[-1]:
                    self.indents.pop()
                    self._pending_tokens.append(
                        Token(TokenType.DEDENT, "", Position(self.line, self.filename))
                    )

            return Token(
                TokenType.NEWLINE, "\n", Position(self.line - 1, self.filename)
            )

        if self.code[self.pos] == "#":
            while self.pos < self.length and self.code[self.pos] != "\n":
                self.pos += 1
            return self.next_token()

        for kind, pattern in TOKEN_REGEX:
            match = pattern.match(self.code, self.pos)
            if match:
                value = match.group(0)
                tok_pos = Position(self.line, self.filename)
                self.pos = match.end()
                if kind in (TokenType.WHITESPACE, TokenType.COMMENT):
                    return self.next_token()  # skip
                return Token(kind, value, tok_pos)

        fbusl_error(
            f"Unexpected character: {self.code[self.pos]}",
            Position(self.line, self.filename),
        )
        self.pos += 1
        return self.next_token()

    def tokenize(self) -> List[Token]:
        tokens = []
        while self.pos < self.length or self._pending_tokens:
            tok = self.next_token()
            if tok:
                tokens.append(tok)

        while len(self.indents) > 1:
            self.indents.pop()
            tokens.append(
                Token(TokenType.DEDENT, "", Position(self.line, self.filename))
            )

        return tokens


class Parser:
    def __init__(self, tokens: list[Token]):
        self.tokens = tokens
        self.position = 0

        self.max_repeat_tokens = 5

    def peek(self) -> Token:
        """Return the current token without consuming it."""
        if self.position < len(self.tokens):
            return self.tokens[self.position]

        return Token("EOF", "", Position(self.position))

    def consume(self) -> Token:
        """Consume and return the current token."""
        tok = self.peek()
        self.position += 1
        return tok

    def expect(self, kind: str, value=None) -> Token:
        """
        Consume the current token if it matches the expected kind (and optionally value),
        otherwise raise an error.
        """
        tok = self.peek()
        if tok.kind != kind or (value is not None and tok.value != value):
            msg = f"Expected token {kind}"
            if value is not None:
                msg += f" with value {value}"
            msg += f', but got {tok.kind} ("{tok.value}")'
            fbusl_error(msg, self.get_current_pos())
        self.position += 1
        return tok

    def get_current_pos(self) -> Position:
        """Return the Position object of the current token."""
        tok = self.peek()
        return tok.pos

    def parse(self):
        last_tokens: list[Token] = []
        ast: list[ASTNode] = []

        while self.position < len(self.tokens):
            token = self.peek()
            node = self.parse_next()

            if node is None:
                self.consume()
                continue

            if isinstance(node, list):
                ast.extend(node)
            else:
                ast.append(node)

            last_tokens.append(token)

            if len(last_tokens) > self.max_repeat_tokens:
                last_tokens.pop(0)

            if len(last_tokens) > self.max_repeat_tokens and all(
                tok.kind == token.kind and tok.value == token.value
                for tok in last_tokens
            ):
                fbusl_error(
                    f"Too many repeats of token: {token}", self.get_current_pos()
                )

        return ast

    def parse_next(self):
        token = self.peek()

        if token.kind == TokenType.KEYWORD:
            if token.value == "def":
                pass

        if token.kind == TokenType.DECORATOR:
            return self.parse_decorator()

        if token.kind == TokenType.KEYWORD:
            if token.value == "def":
                return self.parse_function_def()

        if token.kind == TokenType.IDENT:
            return self.parse_identifier()

    def parse_type(self) -> dict:
        name = self.expect(TokenType.IDENT).value

        if self.peek().value == "[":

            self.expect(TokenType.SYMBOL, "[")
            length = self.expect(TokenType.INT).value
            self.expect(TokenType.SYMBOL, "]")

            return {"name": "array", "data": {"length": length, "base_type": name}}
        else:
            return {"name": name}

    def parse_function_def(self) -> FunctionDef:
        position = self.get_current_pos()
        self.expect(TokenType.KEYWORD, "def")

        name = self.expect(TokenType.IDENT).value
        self.expect(TokenType.SYMBOL, "(")

        params = []

        first = True
        while self.peek().value != ")":
            if not first:
                self.expect(TokenType.SYMBOL, ",")

            param_name = self.expect(TokenType.IDENT)
            param_type, param_qualifier = self.parse_type_and_qualifier()

            if param_qualifier not in ["in", "inout", "out", None]:
                fbusl_error(
                    'Function parameters can only have qualifiers of "in", "out", and "inout".',
                    position,
                )

            params.append(
                FunctionParam(param_name, param_type, param_qualifier, position)
            )

            first = False

        self.expect(TokenType.SYMBOL, ")")

        return_type = None
        if self.peek().kind == TokenType.ARROW:
            self.expect(TokenType.ARROW)
            return_type = self.parse_type()

        self.expect(TokenType.SYMBOL, ":")
        self.expect(TokenType.NEWLINE)
        self.expect(TokenType.INDENT)

        body = []

        while self.peek().kind != TokenType.DEDENT:
            node = self.parse_next()
            if node is None:
                self.consume()  # skip unknown or blank token
                continue
            print(node)

            if isinstance(node, list):
                body.extend(node)
            else:
                body.append(node)

            if self.peek().kind == TokenType.NEWLINE:
                self.consume()

        self.expect(TokenType.DEDENT)

        return FunctionDef(name, body, return_type, params)

    def parse_decorator(self) -> ASTNode:
        decorator_type = self.expect(TokenType.DECORATOR).value
        self.expect(TokenType.NEWLINE)

        decorators = []

        while self.peek().kind == TokenType.IDENT:
            position = self.get_current_pos()
            if decorator_type in ["@uniform", "@input", "@output"]:

                name = self.expect(TokenType.IDENT).value
                var_type, qualifier = self.parse_type_and_qualifier()

                dec_class_map = {
                    "@output": Output,
                    "@input": Input,
                    "@uniform": Uniform,
                }
                decorators.append(
                    dec_class_map[decorator_type](name, var_type, qualifier, position)
                )

            else:
                name = self.expect(TokenType.IDENT)

                self.expect(TokenType.OPERATOR, "=")

                val = self.parse_expression()

                decorators.append(Define(name, val, position))

            self.expect(TokenType.NEWLINE)

        return decorators

    def parse_type_and_qualifier(self) -> tuple[dict, str | None]:
        self.expect(TokenType.SYMBOL, ":")

        prec = None
        if self.peek().kind == TokenType.qualifier:
            prec = self.expect(TokenType.qualifier).value

        node_type = self.parse_type()

        return node_type, prec

    def parse_identifier(self) -> ASTNode:
        node = Identifier(self.consume().value, pos=self.get_current_pos())

        while True:
            tok = self.peek()

            if tok.kind == TokenType.SYMBOL and tok.value == ".":
                self.consume()
                member_tok = self.expect(TokenType.IDENT)
                node = MemberAccess(node, member_tok.value, pos=member_tok.pos)

            elif tok.kind == TokenType.SYMBOL and tok.value == "[":
                self.consume()
                index_expr = self.parse_expression()
                self.expect(TokenType.SYMBOL, "]")
                node = ArrayAccess(node, index_expr, pos=self.get_current_pos())

            elif tok.kind == TokenType.SYMBOL and tok.value == "(":
                self.consume()
                args = self.parse_call_args()
                self.expect(TokenType.SYMBOL, ")")
                node = FuncCall(node, args, pos=self.get_current_pos())

            elif tok.kind == TokenType.OPERATOR and tok.value == "=":
                self.consume()
                expr = self.parse_expression()
                node = Setter(node, expr, self.get_current_pos())

            elif tok.kind == TokenType.SYMBOL and tok.value == ":":
                self.parse_var_decl(node.value)

            else:
                break

        return node

    def parse_var_decl(self, name=None) -> ASTNode:
        pos = self.get_current_pos()

        if name == None:
            name = self.expect(TokenType.IDENT).value

        node_type, qualifier = self.parse_type_and_qualifier()

        value = None
        if self.peek().kind == TokenType.OPERATOR and self.peek().value == "=":
            self.consume()
            value = self.parse_expression()

        return VarDecl(
            name=name, var_type=node_type, value=value, qualifier=qualifier, pos=pos
        )

    def parse_literal(self) -> ASTNode:
        tok = self.consume()
        if tok.kind == TokenType.INT:
            return Literal(int(tok.value), pos=tok.pos)
        elif tok.kind == TokenType.FLOAT:
            return Literal(float(tok.value), pos=tok.pos)
        elif tok.kind == TokenType.KEYWORD and tok.value in ["true", "false"]:
            return Literal(tok.value == "true", pos=tok.pos)
        else:
            fbusl_error(f"Invalid literal: {tok.value}", tok.pos)

    def parse_identifier_expr(self) -> ASTNode:
        node = Identifier(self.consume().value)

        while True:
            tok = self.peek()

            if tok.kind == TokenType.SYMBOL and tok.value == ".":
                self.consume()
                member_tok = self.expect(TokenType.IDENT)
                node = MemberAccess(node, member_tok.value)

            elif tok.kind == TokenType.SYMBOL and tok.value == "[":
                self.consume()
                index_expr = self.parse_expression()
                self.expect(TokenType.SYMBOL, "]")
                node = ArrayAccess(node, index_expr)

            elif tok.kind == TokenType.SYMBOL and tok.value == "(":
                self.consume()
                args = self.parse_call_args()
                self.expect(TokenType.SYMBOL, ")")
                node = FuncCall(node, args)

            else:
                break

        return node

    def parse_call_args(self) -> List[ASTNode]:
        args = []
        tok = self.peek()
        if tok.kind == TokenType.SYMBOL and tok.value == ")":
            return args

        while True:
            arg = self.parse_expression()
            args.append(arg)
            tok = self.peek()
            if tok.kind == TokenType.SYMBOL and tok.value == ",":
                self.consume()
            else:
                break
        return args

    def parse_expression(self, precedence=0) -> ASTNode:
        left = self.parse_primary()

        while True:
            tok = self.peek()
            if tok.kind != TokenType.OPERATOR:
                break
            op = tok.value
            op_prec = self.get_operator_precedence(op)
            if op_prec < precedence:
                break
            self.consume()
            right = self.parse_expression(op_prec + 1)
            left = BinOp(left, op, right)

        return left

    def parse_primary(self) -> ASTNode:
        tok = self.peek()

        if tok.kind == TokenType.IDENT:
            return self.parse_identifier()

        elif tok.kind in (TokenType.INT, TokenType.FLOAT, TokenType.BOOL):
            return self.parse_literal()

        elif tok.kind == TokenType.SYMBOL and tok.value == "(":
            self.consume()  # consume "("
            expr = self.parse_expression()
            self.expect(TokenType.SYMBOL, ")")
            return expr

        elif tok.kind == TokenType.OPERATOR and tok.value in ("+", "-", "not"):
            op = self.consume().value
            operand = self.parse_primary()
            return UnaryOp(op, operand, pos=tok.pos)

        else:
            fbusl_error(
                f"Unexpected token in expression: {tok.kind} ({tok.value})", tok.pos
            )

    def get_operator_precedence(self, op: str) -> int:
        PRECEDENCE = {
            "*": 10,
            "/": 10,
            "%": 10,
            "+": 9,
            "-": 9,
            "<<": 8,
            ">>": 8,
            "<": 7,
            "<=": 7,
            ">": 7,
            ">=": 7,
            "==": 6,
            "!=": 6,
            "&": 5,
            "^": 4,
            "|": 3,
            "and": 2,
            "or": 1,
            "=": 0,
        }
        return PRECEDENCE.get(op, -1)



from FreeBodyEngine.graphics import fbusl
from FreeBodyEngine import get_main
from typing import Literal
import sys
from enum import Enum, auto


class ShaderType:
    VERTEX = auto()
    FRAGMENT = auto()
    COMPUTE = auto()
    GEOMETRY = auto()


def compile(source):
    lexer = fbusl.parser.Lexer(source)
    tokens = lexer.tokenize()

    parser = fbusl.parser.Parser(tokens)
    parser.parse()




from FreeBodyEngine.graphics.fbusl.ast_nodes import *
from FreeBodyEngine.graphics.fbusl import fbusl_error

class GenericVar:
    def __init__(self, name: Identifier | str, type: str, value=None):
        # Accept either an Identifier node or string for name
        if isinstance(name, Identifier):
            self.name = name.name
            self.ident = name
        else:
            self.name = name
            self.ident = Identifier(None, name)
        self.type = type
        self.value = value

    def __repr__(self):
        return "Variable"

class Var(GenericVar):
    pass

class Uniform(GenericVar):
    def __repr__(self):
        return "Uniform"

class Input(Uniform):
    def __repr__(self):
        return "Input"

class Output(Uniform):
    def __repr__(self):
        return f"Output({self.name} type={self.type})"

class Function:
    def __init__(self, name: Identifier | str, return_type, params):
        if isinstance(name, Identifier):
            self.name = name.name
            self.ident = name
        else:
            self.name = name
            self.ident = Identifier(None, name)
        self.params = params  # dict param_name -> type
        self.return_type = return_type

    def __repr__(self):
        return "Function"

class Struct:
    def __init__(self, name: str, attributes: list[dict[str, Type]]):
        self.name = Identifier(None, name)
        self.attributes = attributes

    def __repr__(self):
        return "Struct"

class FuncCall(Node):
    def __init__(self, pos, name: Identifier | str, args):
        super().__init__(pos)
        self.name = name
        self.args = args

class StructCall(Node):
    def __init__(self, pos, name: Identifier | str, args, ismat):
        super().__init__(pos)
        self.ismat = ismat
        self.name = name
        self.args = args

class Definition:
    def __init__(self, name: Identifier | str, val):
        if isinstance(name, Identifier):
            self.name = name.name
            self.ident = name
        else:
            self.name = name
            self.ident = Identifier(None, name)
        self.val = val

    def __repr__(self):
        return "Definition"

class Scope:
    def __init__(self, parent=None, file_path=None):
        self.symbols: dict[str, GenericVar] = {}
        self.parent: Scope | None = parent
        self.file_path = file_path

    def define(self, var: GenericVar):
        if var.name in self.symbols:
            raise SyntaxError(f"{var.__repr__()} '{var.name}' already exists in current scope.")
        self.symbols[var.name] = var

    def lookup(self, name: str, pos) -> GenericVar:
        if name in self.symbols:
            return self.symbols[name]
        if self.parent:
            return self.parent.lookup(name, pos)
        fbusl_error(f"'{name}' not defined", pos, self.file_path)

    def __contains__(self, name: str):
        return name in self.symbols or (self.parent and name in self.parent)

class SemanticAnalyser:
    def __init__(self, tree: Tree, builtins: dict, shader_type, file_path=None):
        self.tree = tree
        self.shader_type = shader_type
        self.global_scope = Scope(file_path=file_path)
        self.current_scope = self.global_scope
        self.functions: dict[str, Function] = {}
        self.structs: dict[str, StructDecl] = {}

        self.file_path = file_path
        self.create_builtins(builtins)

    def create_builtins(self, builtins: dict):
        vars = builtins.get('vars', {})
        outputs = builtins.get('outputs', {})

        for output_name, output_info in outputs.items():
            self.global_scope.define(Output(output_name, output_info.get('type')))
            
        for var_name, var_type in vars.items():
            self.global_scope.define(Var(var_name, var_type))
        
        float_type = Type(0, Identifier(0, "float"))
        int_type = Type(0, Identifier(0, "int"))

        self.functions['round'] = Function('round', Type(0, Identifier(0, float_type)), {"x": float_type})
        self.functions['clamp'] = Function('clamp', Type(0, Identifier(0, float_type)), {"x": float_type})
        self.functions['texture'] = Function('texture', Type(0, Identifier(0, 'vec4')), {"sampler": Type(0, Identifier(0, 'sampler2D')), "sample_pos": Type(0, Identifier(0, 'vec2'))})
        
        self.global_scope.define(Output("VERTEX_POSITION", Type(0, Identifier(0, 'vec4'))))
        if self.shader_type == "vert":
            self.global_scope.define(Output("INSTANCE_ID", Type(0, Identifier(0, 'int'))))

        # Vector structs
        vec2_type = Type(0, Identifier(0, 'vec2'))
        self.structs['vec2'] = StructDecl(None, vec2_type, [
            StructField(None, 'x', float_type, None),
            StructField(None, 'y', float_type, None)
        ])

        vec3_type = Type(0, Identifier(0, 'vec3'))
        self.structs['vec3'] = StructDecl(None, vec3_type, [
            StructField(None, 'x', float_type, None),
            StructField(None, 'y', float_type, None),
            StructField(None, 'z', float_type, None)
        ])

        vec4_type = Type(0, Identifier(0, 'vec4'))
        self.structs['vec4'] = StructDecl(None, vec4_type, [
            StructField(None, 'x', float_type, None),
            StructField(None, 'y', float_type, None),
            StructField(None, 'z', float_type, None),
            StructField(None, 'w', float_type, None)
        ])

        # Int vectors
        self.structs['ivec2'] = StructDecl(None, Type(0, Identifier(0, 'ivec2')), [
            StructField(None, 'x', int_type, None),
            StructField(None, 'y', int_type, None)
        ])
        self.structs['ivec3'] = StructDecl(None, Type(0, Identifier(0, 'ivec3')), [
            StructField(None, 'x', int_type, None),
            StructField(None, 'y', int_type, None),
            StructField(None, 'z', int_type, None)
        ])
        self.structs['ivec4'] = StructDecl(None, Type(0, Identifier(0, 'ivec4')), [
            StructField(None, 'x', int_type, None),
            StructField(None, 'y', int_type, None),
            StructField(None, 'z', int_type, None),
            StructField(None, 'w', int_type, None)
        ])

        # Matrices  
        self.structs['mat2'] = StructDecl(None, Type(0, Identifier(0, 'mat2')), [
            StructField(None, 'col0', vec2_type, None),
            StructField(None, 'col1', vec2_type, None)
        ])
        self.structs['mat3'] = StructDecl(None, Type(0, Identifier(0, 'mat3')), [
            StructField(None, 'col0', vec3_type, None),
            StructField(None, 'col1', vec3_type, None),
            StructField(None, 'col2', vec3_type, None)
        ])
        self.structs['mat4'] = StructDecl(None, Type(0, Identifier(0, 'mat4')), [
            StructField(None, 'col0', vec4_type, None),
            StructField(None, 'col1', vec4_type, None),
            StructField(None, 'col2', vec4_type, None),
            StructField(None, 'col3', vec4_type, None)
        ])

    def analyse(self):
        for node in self.tree.children:
            if isinstance(node, FuncDecl):
                self.check_for_func(node)
            if isinstance(node, StructDecl):
                self.check_for_struct(node)

        for node in self.tree.children:
            self.analyse_node(node)

        return self.tree

    def check_for_func(self, node: FuncDecl):
        params = {}
        for param in node.params:
            params[param.name] = param.type
        self.functions[node.name.name] = Function(node.name, node.return_type, params)

    def check_for_struct(self, node: StructDecl):
        attributes = {}
        for attr in node.methods:
            attributes[attr.name.name] = attr.type
        self.structs[node.name.name] = Struct(node.name, attributes)


    def analyse_node(self, node: Node):
        if isinstance(node, FuncDecl):
            self.analyse_function(node)
        elif isinstance(node, (VarDecl, UniformDecl, InputDecl, OutputDecl, Define)):
            self.analyse_var(node)
        elif isinstance(node, StructDecl):
            self.analyse_struct(node)
        elif isinstance(node, (Set, SetMethod)):
            self.analyse_setter(node)
        elif isinstance(node, Expression):
            self.analyse_expression(node)
        elif isinstance(node, Return):
            self.analyse_node(node.expr)


    def analyse_function(self, node: FuncDecl):
        function_scope = Scope(self.global_scope)
        self.current_scope = function_scope

        for param in node.params:
            param_var = Var(param.name, param.type)
            self.current_scope.define(param_var)

        for stmt in node.body:
            if isinstance(stmt, Return):
                return_type = self.get_node_type(stmt)
                if return_type != node.return_type:
                    fbusl_error(
                        f"Function '{node.name.name}' should return '{node.return_type}', not '{return_type}'",
                        node.pos, self.file_path
                    )
            self.analyse_node(stmt)

        self.current_scope = self.global_scope

    def analyse_var(self, node: Node):
        var_name = node.name.name if hasattr(node, "name") and isinstance(node.name, Identifier) else node.name
        if var_name in self.current_scope:
            fbusl_error(f"Variable '{var_name}' already defined in scope", node.pos, self.file_path)

        if isinstance(node, Define):
            value_type = self.get_node_type(node.val)
            var = Definition(node.name, node.val)
            var.type = value_type
        else:
            var_class = {
                VarDecl: Var,
                UniformDecl: Uniform,
                InputDecl: Input,
                OutputDecl: Output,
            }.get(type(node), Var)
            var = var_class(node.name, node.type)
            
        self.current_scope.define(var)

    def analyse_setter(self, node: Node):
        if isinstance(node.ident, MethodIdentifier):
            struct_var = self.current_scope.lookup(node.ident.struct, node.pos)
            struct_type_name = struct_var.type
            if struct_type_name not in self.structs:
                fbusl_error(f"'{struct_type_name}' is not a struct type", node.pos, self.file_path)
            struct_def = self.structs[struct_type_name]

            if node.ident.method_name not in struct_def.fields:
                fbusl_error(f"Struct '{struct_type_name}' has no field '{node.ident.method_name}'", node.pos, self.file_path)

            expected_type = struct_def.fields[node.ident.method_name]
            value_type = self.get_node_type(node.value)
            if value_type != expected_type:
                
                fbusl_error(f"Type mismatch: Cannot assign {value_type} to {expected_type}", node.pos, self.file_path)

        else:
            var = self.current_scope.lookup(node.ident.name, node.pos)
            if isinstance(var, (Output, Var)):
                value_type = self.get_node_type(node.value)
                if var.type != value_type:
                    fbusl_error(f"Type mismatch: Cannot assign {value_type} to {var.type}", node.pos, self.file_path)
            else:
                fbusl_error(f"Cannot assign to {var.__repr__()}", node.pos, self.file_path)

    def analyse_struct(self, node: StructDecl):
        struct_name = node.name.name if isinstance(node.name, Identifier) else node.name
        if struct_name in self.structs:
            fbusl_error(f"Struct '{struct_name}' already declared.", node.pos, self.file_path)

        field_map = {}
        for method in node.methods:
            if method.name in field_map:
                fbusl_error(f'Struct "{struct_name}" already has field "{method.name}".', method.pos, self.file_path)
            field_map[method.name] = method.type

        node.fields = field_map
        self.structs[struct_name] = node

    def analyse_expression(self, node: Expression):
        if isinstance(node, BinOp):
            left = self.get_node_type(node.left)
            right = self.get_node_type(node.right)
            if left != right:
                fbusl_error(f"Operator '{node.op}' not supported between types '{left}' and '{right}'", node.pos, self.file_path)

    def get_node_type(self, node: Node)->Type:
        if isinstance(node, Expression):
            return self.get_expression_type(node)
        elif isinstance(node, Return):
            return self.get_node_type(node.expr)
        elif isinstance(node, Int):
            return "int"
        elif isinstance(node, Float):
            return "float"
        elif isinstance(node, MethodIdentifier):
            return self.get_field_type(node)
        elif isinstance(node, Identifier):
            var = self.current_scope.lookup(node.name, node.pos)
            return var.type
        elif isinstance(node, Get):
            if not node.array_access:
                var = self.current_scope.lookup(node.name.name, node.pos)
                return var.type
            else:
                var = self.current_scope.lookup(node.name.name, node.pos)
                return var.type.base_type()
        elif isinstance(node, MethodGet):
            if not node.array_access:
                struct = self.structs.get(node.struct_name)
                return struct.methods.get(node.name.name)
            
            else:
                struct = self.structs.get(node.struct_name)
                return struct.methods.get(node.name.name).base_type()

        
        elif isinstance(node, TypeCast):
            return node.type
        elif isinstance(node, TernaryExpression):
            return self.get_node_type(node.left)
        elif isinstance(node, Call):
            call_name = node.name.name if isinstance(node.name, Identifier) else node.name
            if call_name in self.functions:
                return self.functions[call_name].return_type
            if call_name in self.structs:
                
                return Type(node.pos, Identifier(node.pos, call_name))
            
            fbusl_error(f"'{call_name}' not defined", node.pos, self.file_path)
        else:
            fbusl_error(f"Cannot determine type of node {type(node).__name__}", getattr(node, 'pos', None), self.file_path)


    def get_field_type(self, node: MethodGet):
        struct_var = self.current_scope.lookup(node.struct, node.pos)
        struct_type_name = struct_var.type

        if struct_type_name not in self.structs:
            fbusl_error(f"'{struct_type_name}' is not a struct type", node.pos, self.file_path)
        
        struct_def = self.structs[struct_type_name]
        
        if node.method_name not in struct_def.fields:
            fbusl_error(f"Struct '{struct_type_name}' has no field '{node.method_name}'", node.pos, self.file_path)
        
        return struct_def.fields[node.method_name]

    def get_expression_type(self, node: Expression):
        if isinstance(node, BinOp):
            left = self.get_node_type(node.left)
            right = self.get_node_type(node.right)
            for i in range(2, 5):
                if left.name.name == f'mat{i}' and right.name.name == f'vec{i}':
                    return Type(node.pos, Identifier(node.pos, f'vec{i}'))
            
            if left == right:
                return left
            
            fbusl_error(f"Operator '{node.op}' used between mismatched types '{left}' and '{right}'", node.pos, self.file_path)
        
        elif isinstance(node, FuncCall):
            call_name = node.name.name if isinstance(node.name, Identifier) else node.name
            if call_name in self.functions:
                return self.functions[call_name].return_type
            fbusl_error(f"Function '{call_name}' not defined", node.pos, self.file_path)

        elif isinstance(node, StructCall):
            struct_name = node.name.name if isinstance(node.name, Identifier) else node.name
            if struct_name in self.structs:
                return struct_name
            fbusl_error(f"Struct '{struct_name}' not defined", node.pos, self.file_path)

        elif isinstance(node, Identifier):
            var = self.current_scope.lookup(node.name, node.pos)
            return var.type

        elif isinstance(node, MethodIdentifier):
            return self.get_field_type(node)

        elif isinstance(node, Int):
            return "int"

        elif isinstance(node, Float):
            return "float"

        elif isinstance(node, Bool):
            return "bool"

        elif isinstance(node, TypeCast):
            return node.type

        elif isinstance(node, TernaryExpression):
            left_type = self.get_node_type(node.left)
            right_type = self.get_node_type(node.right)
            if left_type != right_type:
                fbusl_error(f"Ternary branches have mismatched types '{left_type}' and '{right_type}'", node.pos, self.file_path)
            return left_type

        else:
            fbusl_error(f"Unknown expression type: {type(node).__name__}", getattr(node, 'pos', None), self.file_path)



class Tree:
    def __init__(self):
        self.children = []

    def build_tree_lines(self, node_repr, prefix="", is_last=True):
        label = node_repr[0]
        lines = [prefix + ("└── " if is_last else "├── ") + label]
        children = node_repr[1:]
        for i, child in enumerate(children):
            is_child_last = i == len(children) - 1
            if isinstance(child, Node):
                lines += self.build_tree_lines(child.get_debug(),
                    prefix + ("    " if is_last else "│   "),
                    is_child_last
                )
            else:
                lines.append(
                    prefix + ("    " if is_last else "│   ") +
                    ("└── " if is_child_last else "├── ") + str(child)
                )
        return lines

    def __str__(self):
        if not self.children:
            return ""
        lines = []
        for i, child in enumerate(self.children):
            is_last = i == len(self.children) - 1
            lines += self.build_tree_lines(child.get_debug(), "", is_last)
        return "\n".join(lines)

class Node:
    def __init__(self, pos: int):
        self.pos = pos

class Expression(Node):
    def __init__(self, pos: int):
        super().__init__(pos)

class Identifier(Node):
    def __init__(self, pos, name):
        super().__init__(pos)
        self.name = name
        
    def get_debug(self):
        return (f"Identifier('{self.name}')",)

class MethodIdentifier(Identifier):
    def __init__(self, pos, struct, method_name):
        super().__init__(pos, f"{struct}.{method_name}")
        self.struct = struct
        self.method_name = method_name
    def get_debug(self):
        return (f"MethodIdentifier('{self.method_name}')", self.struct)

class Buffer(Node):
    def __init__(self, pos, name: Identifier, fields: list):
        super().__init__(pos)
        self.name = name
        self.fields = fields
    
    def get_debug(self):
        return (f"Buffer", self.name, self.fields)

class Type(Node):
    def __init__(self, pos, name: Identifier, length=1):
        super().__init__(pos)
        self.name = name
        self.length = length
    
    def base_type(self):
        return Type(self.pos, self.name)

    def get_debug(self):
        return (f"Type('{self.name.name}'), len={self.length}",)

    def __eq__(self, other):
        if isinstance(other, Type):
            return other.name.name == self.name.name and other.length == self.length

    def __str__(self):
        if not self.length > 1:
            return f"Type({self.name.name})"
        else:
            return f"Type({self.name.name})[{self.length}]"


class BufferField(Node):
    def __init__(self, pos, name: Identifier, type: str):
        super().__init__(pos)
        self.name = name
        self.type = type

    def get_debug(self):
        return (f"BufferField {self.type}", self.name)


class BinOp(Expression):
    def __init__(self, pos: int, left, op, right):
        super().__init__(pos)
        self.left = left
        self.op = op
        self.right = right

    def get_debug(self):
        return (f"BinOp('{self.op}')", self.left, self.right)

class VarDecl(Node):
    def __init__(self, pos, name, type, val, precision):
        super().__init__(pos)
        self.name = name
        self.type = type
        self.val = val
        self.precision = precision
    def get_debug(self):
        return (f"VariableDeclaration", self.name, self.type, self.val)

class UniformDecl(Node):
    def __init__(self, pos, name, type, precision):
        super().__init__(pos)
        self.name = name
        self.type = type
        self.precision = precision
    def get_debug(self):
        return (f"UniformDeclaration", self.name, self.type)

class Get(Node):
    def __init__(self, pos, name: Identifier, array_access: bool, array_index=0):
        super().__init__(pos)
        self.name = name
        self.array_access = array_access
        self.array_index = array_index

    def get_debug(self):
        return (f"Get array={self.array_access}", self.name)
    
class MethodGet(Node):
    def __init__(self, pos, struct_name: Node, name: Identifier, array_access: bool, array_index=0):
        super().__init__(pos)
        self.struct_name = struct_name
        self.name = name                
        self.array_access = array_access
        self.array_index = array_index

    def get_debug(self):
        kind = "array" if self.array_access else "field"
        return (f"MethodGet ({kind}) to {self.name.name}", self.struct_name)


class InputDecl(Node):
    def __init__(self, pos, name, type, precision):
        super().__init__(pos)
        self.name = name
        self.type = type
        self.precision = precision
    def get_debug(self):
        return (f"InputDeclaration", self.name, self.type)

class OutputDecl(Node):
    def __init__(self, pos, name, type, precision):
        super().__init__(pos)
        self.name = name
        self.type = type
        self.precision = precision
    def get_debug(self):
        return (f"OutputDeclaration", self.name, self.type)

class Define(Node):
    def __init__(self, pos, name, val):
        super().__init__(pos)
        self.name = name
        self.val = val
    def get_debug(self):
        return (f"Definition", self.name, self.val)

class Set(Node):
    def __init__(self, pos, ident, value):
        super().__init__(pos)
        self.ident = ident
        self.value = value
    def get_debug(self):
        return ("Set", self.ident, self.value)

class SetMethod(Set):
    def __init__(self, pos, identifier, method, value):
        super().__init__(pos, identifier, value)
        self.method = method
    def get_debug(self):
        return (f"SetMethod", self.ident, self.method, self.value)

class Not(Node):
    def __init__(self, pos, node):
        super().__init__(pos)
        self.node = node
    def get_debug(self):
        return ("Not", self.node)

class And(Node):
    def __init__(self, pos, left, right):
        super().__init__(pos)
        self.left = left
        self.right = right
    def get_debug(self):
        return ("And", self.left, self.right)

class Or(Node):
    def __init__(self, pos, left, right):
        super().__init__(pos)
        self.left = left
        self.right = right
    def get_debug(self):
        return ("Or", self.left, self.right)

class IfStatement(Node):
    def __init__(self, pos, condition, body):
        super().__init__(pos)
        self.condition = condition
        self.body = body

    def get_debug(self):
        return ("IfStatement", self.condition, *self.body)

class ElseIfStatement(IfStatement):
    def get_debug(self):
        return ("ElseIfStatement", self.condition, *self.body)

class ElseStatement(Node):
    def __init__(self, pos, body):
        super().__init__(pos)
        self.body = body

    def get_debug(self):
        return ("ElseStatement", *self.body)

class Condition(Node):
    def __init__(self, pos, left, right, comparison):
        super().__init__(pos)
        self.left = left
        self.right = right
        self.comparison = comparison
    def get_debug(self):
        return (f"Condition('{self.comparison}')", self.left, self.right)

class TernaryExpression(Node):
    def __init__(self, pos, left, right, condition):
        super().__init__(pos)
        self.left = left
        self.right = right
        self.condition = condition
    def get_debug(self):
        return ("TernaryExpression", self.left, self.right, self.condition)

class Param(Node):
    def __init__(self, pos, name, type):
        super().__init__(pos)
        self.name = name
        self.type = type
    def get_debug(self):
        return (f"Param", self.name, self.type)

class Return(Node):
    def __init__(self, pos, expr):
        super().__init__(pos)
        self.expr = expr
    def get_debug(self):
        return ("Return", self.expr)

class FuncDecl(Node):
    def __init__(self, pos, name, return_type, params, body):
        super().__init__(pos)
        self.name = name
        self.return_type = return_type
        self.params = params
        self.body = body
    def get_debug(self):
        return (f"Function -> {self.return_type}", self.name, *self.params, *self.body)

class StructField(Node):
    def __init__(self, pos, name, type, precision):
        super().__init__(pos)
        self.name = name
        self.type = type
        self.precision = precision
    def get_debug(self):
        return (f"Field", self.name, self.type)

class Call(Node):
    def __init__(self, pos, name, args):
        super().__init__(pos)
        self.name = name
        self.args = args
    def get_debug(self):
        return (f"Call", self.name, *self.args)

class Arg(Node):
    def __init__(self, pos, val):
        super().__init__(pos)
        self.val = val
    def get_debug(self):
        return ("Argument", self.val)

class StructDecl(Node):
    def __init__(self, pos, name, methods):
        super().__init__(pos)
        self.name = name
        self.methods = methods
    def get_debug(self):
        return (f"Struct", self.name, *self.methods)

class TypeCast(Node):
    def __init__(self, pos, target, type):
        super().__init__(pos)
        self.type = type
        self.target = target
    def get_debug(self):
        return ("TypeCast", self.target, self.type)

# Literals
class Int(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Integer({self.value})",)

class Float(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Float({self.value})",)

class Bool(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Boolean({self.value})",)

class Vec2(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Vector2({self.value})",)

class Vec3(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Vector3({self.value})",)

class Vec4(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Vector4({self.value})",)

class Mat2(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Matrix2x2({self.value})",)

class Mat3(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Matrix3x3({self.value})",)

class Mat4(Expression):
    def __init__(self, pos, value):
        super().__init__(pos)
        self.value = value
    def get_debug(self):
        return (f"Matrix4x4({self.value})",)


from typing import List, Optional, Dict, Union
from FreeBodyEngine.graphics.fbusl import Position
class ASTNode:
    def __init__(self, pos: Position = Position()):
        self.pos: Position = pos
        self.children: List['ASTNode'] = []
        self.type: Optional[dict] = None
        self.scope: Optional[Dict[str, str]] = None
        self.metadata: Dict[str, Union[str, int]] = {}
        self.value: Optional[Union[str, int, float]] = None

    def add_child(self, node: 'ASTNode'):
        self.children.append(node)

    def __repr__(self):
        return f"{self.__class__.__name__}(pos={self.pos})"


class Program(ASTNode):
    def __init__(self, pos: Position = Position()):
        super().__init__(pos)
        self.functions: List['FunctionDef'] = []
        self.outputs: List['OutputVar'] = []

    def __repr__(self):
        return f"Program(functions={self.functions}, outputs={self.outputs}, pos={self.pos})"


class FunctionDef(ASTNode):
    def __init__(self, name: str, body: List[ASTNode], return_type: str, params: List['FunctionParam'], overloads: Optional[List['FunctionParam']] = [], pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.type = return_type
        self.params = params
        self.body: List[ASTNode] = body
        for node in self.body:
            self.add_child(node)
        self.overloads: List[Dict] = overloads

    def __repr__(self):
        return (f"FunctionDef(name={self.name}, return_type={self.type}, "
                f"params={self.params}, body={self.body}, overloads={self.overloads}, pos={self.pos})")

class FunctionParam(ASTNode):
    def __init__(self, name: str, var_type: dict, qualifier: str | None, pos: Position = Position()):
        super().__init__(pos)
        self.qualifier = qualifier
        self.name = name
        self.var_type = var_type
        

class VarDecl(ASTNode):
    def __init__(self, name: str, var_type: dict, value: Optional[ASTNode] = None, pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.type = var_type
        self.value = value
        if value:
            self.add_child(value)

    def __repr__(self):
        return f"VarDecl(name={self.name}, type={self.type}, value={self.value}, pos={self.pos})"

class Setter(ASTNode):
    def __init__(self, node: ASTNode, to: ASTNode, pos: Position = Position()):
        super().__init__(pos)
        self.node = node
        self.to = to
        self.add_child(to)
    
    def __repr__(self):
        return f"Setter(node={self.node}, to={self.to}, pos={self.pos})"

class Output(ASTNode):
    def __init__(self, name: str, var_type: dict, qualifier: str, pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.type = var_type
        self.qualifier = qualifier

    def __repr__(self):
        return f"Output(name={self.name}, type={self.type}, qualifier={self.qualifier}, pos={self.pos})"


class Input(ASTNode):
    def __init__(self, name: str, var_type: dict, qualifier: str, pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.type = var_type
        self.qualifier = qualifier

    def __repr__(self):
        return f"Input(name={self.name}, type={self.type}, qualifier={self.qualifier}, pos={self.pos})"


class Uniform(ASTNode):
    def __init__(self, name: str, var_type: dict, qualifier: str, pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.type = var_type
        self.qualifier = qualifier

    def __repr__(self):
        return f"Uniform(name={self.name}, type={self.type}, qualifier={self.qualifier}, pos={self.pos})"


class Define(ASTNode):
    def __init__(self, name: str, value: ASTNode, pos: Position = Position()):
        super().__init__(pos)
        self.name = name
        self.value = value

    def __repr__(self):
        return f"Define(name={self.name}, value={self.value}, pos={self.pos})"


class BinOp(ASTNode):
    def __init__(self, left: ASTNode, op: str, right: ASTNode, pos: Position = Position()):
        super().__init__(pos)
        self.left = left
        self.right = right
        self.op = op
        self.add_child(left)
        self.add_child(right)

    def __repr__(self):
        return f"BinOp(op={self.op}, left={self.left}, right={self.right}, pos={self.pos})"


class UnaryOp(ASTNode):
    def __init__(self, op: str, operand: ASTNode, pos: Position = Position()):
        super().__init__(pos)
        self.op = op
        self.operand = operand
        self.add_child(operand)

    def __repr__(self):
        return f"UnaryOp(op={self.op}, operand={self.operand}, pos={self.pos})"


class MemberAccess(ASTNode):
    def __init__(self, base: ASTNode, member: str, pos: Position = Position()):
        super().__init__()
        self.base = base
        self.member = member
        self.pos = pos
        self.add_child(base)

    def __repr__(self):
        return f"MemberAccess(base={self.base}, member={self.member}, pos={self.pos})"


class ArrayAccess(ASTNode):
    def __init__(self, base: ASTNode, index: ASTNode, pos: Position = Position()):
        super().__init__()
        self.base = base
        self.index = index
        self.pos = pos
        self.add_child(base)
        self.add_child(index)

    def __repr__(self):
        return f"ArrayAccess(base={self.base}, index={self.index}, pos={self.pos})"


class FuncCall(ASTNode):
    def __init__(self, func: ASTNode, args: List[ASTNode], pos: Position = Position()):
        super().__init__()
        self.func = func
        self.args = args
        self.pos = pos
        self.add_child(func)
        for arg in args:
            self.add_child(arg)

    def __repr__(self):
        return f"FuncCall(func={self.func}, args={self.args}, pos={self.pos})"


class Literal(ASTNode):
    def __init__(self, value: Union[int, float, bool], pos: Position = Position()):
        super().__init__(pos)
        self.value = value
        if isinstance(value, int):
            self.type = "int"
        elif isinstance(value, float):
            self.type = "float"
        elif isinstance(value, bool):
            self.type = "bool"

    def __repr__(self):
        return f"Literal(value={self.value}, type={self.type}, pos={self.pos})"


class Identifier(ASTNode):
    def __init__(self, name: str, pos: Position = Position()):
        super().__init__(pos)
        self.value = name

    def __repr__(self):
        return f"Identifier(name={self.value}, pos={self.pos})"



from FreeBodyEngine.graphics.fbusl.ast_nodes import *
from FreeBodyEngine.utils import abstractmethod
from FreeBodyEngine.graphics.fbusl.semantic import SemanticAnalyser

class Generator:
    """
    The code generator for FBUSL ASTs.
    """
    def __init__(self, tree: Tree, analyser: SemanticAnalyser, buffer_index: int):
        self.tree = tree
        self.analyser = analyser
        self.buffer_index = buffer_index

    @abstractmethod
    def generate(self) -> str:
        pass


from typing import List
import re
from FreeBodyEngine.graphics.fbusl.ast_nodes import *
from FreeBodyEngine.graphics.fbusl import fbusl_error

class Token:
    def __init__(self, kind: str, value, pos: int):
        self.kind = kind
        self.value = value
        self.pos = pos
    def __repr__(self):
        return f"Token({self.kind}, {repr(self.value)})"

TOKEN_TYPES = [
    ("ARROW", r"->"),
    ("DECORATOR", r"@(?:uniform|input|output|define)"),
    ("PRECISION", r"(low|med|flat|high)"),
    ("KEYWORD", r"\b(def|return|struct|if|buffer|elif|else|not|and|or)\b"),
    ("COMPARISON", r"(==|!=|<=|>=|<|>)"),
    ("TYPE", r"\b(void|float|int|bool|sampler2D|sampler3D)\b"),
    ("IDENT", r"[a-zA-Z_][a-zA-Z0-9_]*"),
    ("FLOAT", r"\d+\.\d+"),
    ("INT", r"\d+"),
    ("SYMBOL", r"[{}()[\]:,\.]"),
    ("OPERATOR", r"(\+=|-=|\*=|/=|=|\+|-|\*|/)"),
    ("WHITESPACE", r"[ \t]+"),
    ("COMMENT", r"#.*"),
]

def tokenize(code: str, file_path) -> List[Token]:
    tokens = []
    indent_stack = [0]
    lines = code.splitlines()

    for line_num, line in enumerate(lines, start=1):
        if re.match(r"^\s*$", line) or re.match(r"^\s*#", line):
            continue

        indent_match = re.match(r"[ \t]*", line)
        indent_str = indent_match.group(0)
        indent = len(indent_str.replace('\t', '    '))  # convert tabs to spaces

        if indent > indent_stack[-1]:
            tokens.append(Token("INDENT", indent, pos=line_num))
            indent_stack.append(indent)
        elif indent < indent_stack[-1]:
            while indent < indent_stack[-1]:
                indent_stack.pop()
                tokens.append(Token("DEDENT", indent_stack[-1], pos=line_num))
            if indent != indent_stack[-1]:
                fbusl_error(f"Inconsistent indentation: expected {indent_stack[-1]}, got {indent}", line_num, file_path)

        pos = len(indent_str)
        line_length = len(line)
        while pos < line_length:
            for kind, pattern in TOKEN_TYPES:
                match = re.match(pattern, line[pos:])
                if match:
                    value = match.group(0)
                    if kind == "WHITESPACE":
                        pos += len(value)
                        break
                    elif kind == "COMMENT":
                        pos = line_length 
                        break
                    else:
                        token = Token(kind, value, pos=line_num)
                        tokens.append(token)
                        pos += len(value)
                        break
            else:
                fbusl_error(f"Unexpected character {line[pos]!r}", line_num, file_path)

        tokens.append(Token("NEWLINE", "\\n", pos=line_num))

    while len(indent_stack) > 1:
        indent_stack.pop()
        tokens.append(Token("DEDENT", indent_stack[-1], pos=line_num + 1))

    return tokens

class FBUSLParser:
    """Parses FBUSL code into an AST."""
    def __init__(self, tokens, file_path):
        self.tokens: list[Token] = tokens
        self.index = 0
        self.file_path = file_path

    def peek(self, offset=0) -> Token:
        if not self.index + offset > len(self.tokens) - 1:
            return self.tokens[self.index + offset]
        return 0

    def expect(self, kind: str, value: str = None):
        tok = self.peek()

        if tok == 0:
            if value is None:
                fbusl_error(f"Missing '{kind}'", self.tokens[-1].pos, self.file_path)
            else:
                fbusl_error(f"Missing '{kind}' '{value}'", self.tokens[-1].pos, self.file_path)

        if value is not None:
            if tok.value != value:
                fbusl_error(
                    f"Expected token of kind '{kind}' with value of {value}, instead got '{tok.kind}' '{tok.value}'.",
                    tok.pos, self.file_path)

        if tok.kind != kind:
            fbusl_error(f"Expected token of kind '{kind}', instead got kind '{tok.kind}' '{tok.value}'", tok.pos, self.file_path)

        self.consume()
        return tok

    def consume(self):
        tok = self.peek()
        self.index += 1
        return tok

    def parse(self) -> Tree:
        tree = Tree()
        while self.index < len(self.tokens):
            n = self.parse_next()
            if n == None:
                continue
            tree.children.append(n)

        tree.children = [c for c in tree.children if c is not None]
        return tree

    def parse_next(self) -> Node:
        tok = self.peek()
        if tok.kind == "NEWLINE":
            self.consume()
        if tok.kind == "KEYWORD":
            if tok.value == "def":
                return self.parse_function()
            elif tok.value == "struct":
                return self.parse_struct()
            elif tok.value == 'return':
                return self.parse_return()
            elif tok.value in ['if', 'elif', 'else']:
                return self.parse_if_statement()
            elif tok.value == "buffer":
                return self.parse_buffer()
        elif tok.kind == "TYPE":
            return self.parse_typecast()
        elif tok.kind in ["PRECISION", "IDENT"]:
            return self.parse_var()
        elif tok.kind == "DECORATOR":
            return self.parse_decorator()

    def parse_decorator(self):
        decorator = self.expect('DECORATOR').value
        self.expect('NEWLINE')
        if decorator == "@define":
            return self.parse_definition()
        elif decorator == '@output':
            return self.parse_inout('out')
        elif decorator == '@input':
            return self.parse_inout('in')
        elif decorator == '@uniform':
            return self.parse_uniform()

    def parse_definition(self):
        name = self.expect('IDENT')
        self.expect('OPERATOR', '=')
        val = self.parse_expression()
        return Define(name.pos, Identifier(name.pos, name.value), val)

    def parse_typecast(self):
        type = self.expect("TYPE").value
        self.expect("SYMBOL", "(")
        target = self.parse_expression()
        self.expect("SYMBOL", ")")
        return TypeCast(target.pos, target, type)

    def parse_uniform(self):
        precision = None
        if self.peek().kind == "PRECISION":
            precision = self.expect('PRECISION').value

        name = self.expect('IDENT')
        self.expect('SYMBOL', ':')

        type = self.parse_type()

        return UniformDecl(name.pos, Identifier(name.pos, name.value), type, precision)

    def parse_if_statement(self):
        type = self.expect("KEYWORD")
        condition = None

        if type.value != "else":
            condition = self.parse_expression()

        self.expect('SYMBOL', ':')
        self.expect('NEWLINE')
        self.expect('INDENT')

        body = []
        while self.peek().kind != "DEDENT":
            
            n = self.parse_next()
            if n == None:
                continue
            body.append(n)
            
        self.expect("DEDENT")

        if type.value == 'if':
            return IfStatement(type.pos, condition, body)
        elif type.value == 'elif':
            return ElseIfStatement(type.pos, condition, body)
        else:
            return ElseStatement(type.pos, body)

    def parse_inout(self, inout_type: str):
        precision = None
        if self.peek().kind == "PRECISION":
            precision = self.expect('PRECISION').value

        name = self.expect('IDENT')
        self.expect('SYMBOL', ':')
        
        type = self.parse_type()

        ident = Identifier(name.pos, name.value)
        if inout_type == "in":
            return InputDecl(name.pos, ident, type, precision)
        elif inout_type == "out":
            return OutputDecl(name.pos, ident, type, precision)

    def parse_return(self):
        pos = self.expect("KEYWORD", 'return').pos
        val = self.parse_expression()
        return Return(pos, val)

    def parse_type(self, allow_array=True):
        if self.peek().kind == "TYPE":
            type_name = self.expect('TYPE')
        else:
            type_name = self.expect('IDENT')

        if allow_array:
            if self.peek().kind == "SYMBOL" and self.peek().value == "[":
                self.expect("SYMBOL", "[")
                size_token = self.expect("INT")
                size = int(size_token.value)
                self.expect("SYMBOL", "]")
                return Type(type_name.pos, Identifier(type_name.pos, type_name.value), size)
        
        return Type(type_name.pos, Identifier(type_name.pos, type_name.value))

    def parse_function(self):
        self.expect("KEYWORD", "def")
        name = self.expect("IDENT")
        self.expect("SYMBOL", "(")

        params = []
        if self.peek().kind != "SYMBOL" or self.peek().value != ")":
            params.append(self.parse_param())
            while self.peek().kind == "SYMBOL" and self.peek().value == ",":
                self.consume()
                params.append(self.parse_param())

        self.expect("SYMBOL", ")")
        return_type = None
        if self.peek().kind == "ARROW":
            self.expect("ARROW", "->")
            return_type = self.parse_type(False)

        self.expect("SYMBOL", ":")
        self.expect("NEWLINE")
        self.expect("INDENT")

        body = []
        while self.peek().kind != "DEDENT":
            stmt = self.parse_next()
            if stmt:
                body.append(stmt)

        self.expect("DEDENT")
        return FuncDecl(name.pos, Identifier(name.pos, name.value), return_type, params, body)

    def parse_param(self) -> Param:
        name = self.expect('IDENT')
        self.expect("SYMBOL", ":")
        type = self.parse_type()

        return Param(name.pos, Identifier(name.pos, name.value), type)

    def parse_expression(self):
        expr = self.parse_logical_or()
        if self.peek().kind == "KEYWORD" and self.peek().value == "if":
            self.consume()
            condition = self.parse_logical_or()
            self.expect("KEYWORD", "else")
            false_expr = self.parse_expression()
            return TernaryExpression(expr.pos, expr, false_expr, condition)
        return expr

    def parse_logical_or(self):
        left = self.parse_logical_and()
        while self.peek().kind == "KEYWORD" and self.peek().value == "or":
            self.consume()
            right = self.parse_logical_and()
            left = Or(left.pos, left, right)
        return left

    def parse_logical_and(self):
        left = self.parse_equality()
        while self.peek().kind == "KEYWORD" and self.peek().value == "and":
            self.consume()
            right = self.parse_equality()
            left = And(left.pos, left, right)
        return left

    def parse_equality(self):
        left = self.parse_condition()
        while self.peek().kind == "COMPARISON" and self.peek().value in ["==", "!="]:
            comparison = self.consume().value
            right = self.parse_condition()
            left = Condition(left.pos, left, right, comparison)
        return left

    def parse_condition(self):
        left = self.parse_term()
        while self.peek().kind == "COMPARISON" and self.peek().value in ["<", ">", "<=", ">="]:
            comparison = self.consume().value
            right = self.parse_term()
            left = Condition(left.pos, left, right, comparison)
        return left

    def parse_term(self):
        left = self.parse_factor()
        while self.peek().kind == "OPERATOR" and self.peek().value in ["+", "-"]:
            op = self.consume().value
            right = self.parse_factor()
            left = BinOp(left.pos, left, op, right)
        return left

    def parse_factor(self):
        left = self.parse_unary()
        while self.peek().kind == "OPERATOR" and self.peek().value in ["*", "/"]:
            op = self.consume().value
            right = self.parse_unary()
            left = BinOp(left.pos, left, op, right)
        return left

    def parse_unary(self):
        if self.peek().kind == "KEYWORD" and self.peek().value == "not":
            op_tok = self.consume()
            operand = self.parse_unary()
            return Not(op_tok.pos, operand)
        return self.parse_value()

    def parse_var(self):
        precision = None
        if self.peek().kind == "PRECISION":
            precision = self.expect('PRECISION').value
        name = self.expect('IDENT')

        if self.peek().value == ":":
            return self.parse_var_decl(name, precision)
        elif precision is None:
            return self.parse_var_set(name)
        else:
            fbusl_error('Cannot change precision of defined value', name.pos, self.file_path)

    def parse_var_decl(self, name, precision):
        self.expect("SYMBOL", ":")
        type = self.parse_type()

        val = None
        if self.peek().kind == 'OPERATOR' and self.peek().value == "=":
            self.consume()
            val = self.parse_expression()

        return VarDecl(name.pos, Identifier(name.pos, name.value), type, val, precision)

    def parse_var_set(self, name):
        self.expect("OPERATOR", "=")
        expression = None
        if self.peek().kind != "NEWLINE":
            expression = self.parse_expression()
        return Set(name.pos, Identifier(name.pos, name.value), expression)

    def parse_value(self):
        tok = self.peek()
        if tok.kind == "INT":
            return Int(tok.pos, self.consume().value)
        elif tok.kind == "FLOAT":
            return Float(tok.pos, self.consume().value)
        elif tok.kind == "IDENT":
            return self.parse_get()
        elif tok.kind == "TYPE" and self.peek(1).value == "(":
            return self.parse_typecast()
        else:
            fbusl_error(f"Unexpected token '{tok.value}'", tok.pos, self.file_path)

    def parse_get(self) -> Node:
        node = Identifier(self.peek().pos, self.expect("IDENT").value)

        while self.peek().kind == "SYMBOL" and self.peek().value in [".", "[", "("]:
            if self.peek().value == ".":
                self.expect("SYMBOL", ".")
                field = self.expect("IDENT")
                node = MethodGet(field.pos, node, Identifier(field.pos, field.value), array_access=False)

            elif self.peek().value == "[":
                lbrack = self.expect("SYMBOL", "[")
                index_expr = self.parse_expression()
                self.expect("SYMBOL", "]")

                if isinstance(node, MethodGet):
                    node.array_access = True
                    node.array_index = index_expr
                else:
                    node = Get(lbrack.pos, node, array_access=True, array_index=index_expr)

            elif self.peek().value == "(":
                if isinstance(node, MethodGet):
                    fbusl_error("Cannot call struct field: function calls like foo.bar() are not allowed")
                self.expect("SYMBOL", "(")
                args = []
                while self.peek().value != ")":
                    if self.peek().value == ",":
                        self.consume()
                    args.append(Arg(node.pos, self.parse_expression()))
                self.expect("SYMBOL", ")")
                if not isinstance(node, Identifier):
                    fbusl_error('Call name must be an identifier.')
                node = Call(node.pos, node, args)

        return node

    def parse_field(self) -> StructField:
        precision = None
        if self.peek().kind == "PRECISION":
            precision = self.expect('PRECISION').value

        name = self.expect('IDENT')
        self.expect("SYMBOL", ":")

        type = self.parse_type()

        return StructField(name.pos, Identifier(name.pos, name.value), type, precision)

    
    def parse_buffer(self) -> Buffer:
        self.expect('KEYWORD', 'buffer')
        name = self.expect('IDENT')
        self.expect('SYMBOL', ':')
        self.expect("NEWLINE")

        fields = []
        while self.peek().kind != "DEDENT":
            self.expect('INDENT')
            field_name = self.expect('IDENT')
            self.expect("SYMBOL", ":")
            
            field_type = self.parse_type()
            
            fields.append(BufferField(field_name.pos, Identifier(field_name.pos, field_name.value), field_type))

            self.expect("NEWLINE")

        self.expect("DEDENT")



        return Buffer(name.pos, Identifier(name.pos, name.value), fields)


    def parse_struct(self) -> StructDecl:
        self.expect('KEYWORD', 'struct')
        name = self.expect('IDENT')
        self.expect('SYMBOL', ":")
        self.expect('NEWLINE')
        self.expect('INDENT')

        fields = []
        while self.peek().kind != "DEDENT":
            fields.append(self.parse_field())
            if self.peek().kind == "NEWLINE":
                self.expect('NEWLINE')
        self.expect("DEDENT")
        return StructDecl(name.pos, Identifier(name.pos, name.value), fields)
    
class GLSLParser:
    """Parses GLSL into an AST."""